function rankOf(node) {
    if (node !== undefined) {
        return node.rank;
    }
    return 0;
}
export class ValuedNode {
    static leaf(iRange) {
        return new ValuedNode(undefined, iRange, undefined);
    }
    constructor(l, v, r) {
        this.left = l;
        this.value = v;
        this.right = r;
        this.rank = 0;
        this.update();
    }
    // Basic changes
    setLeft(l) {
        ;
        this.left = l;
        this.update();
    }
    setRight(r) {
        ;
        this.right = r;
        this.update();
    }
    update() {
        this.rank = 1 + Math.max(rankOf(this.left), rankOf(this.right));
    }
    updated() {
        return this.balance();
    }
    // Balancing
    balanceFactor() {
        return rankOf(this.right) - rankOf(this.left);
    }
    isBalanced() {
        const balanceFactor = this.balanceFactor();
        return -2 < balanceFactor && balanceFactor < 2;
    }
    isRightUnbalanced() {
        return this.balanceFactor() >= 2;
    }
    isRightOriented() {
        return this.balanceFactor() === 1;
    }
    isLeftOriented() {
        return this.balanceFactor() <= -2;
    }
    isSlighlyLeftUnbalanced() {
        return this.balanceFactor() === -1;
    }
    rotateLeft() {
        const r = this.right;
        this.setRight(r.left);
        r.setLeft(this);
        return r;
    }
    rotateRight() {
        const l = this.left;
        this.setLeft(l.right);
        l.setRight(this);
        return l;
    }
    balance() {
        let self = this;
        while (self.isRightUnbalanced()) {
            if (self.right.isSlighlyLeftUnbalanced()) {
                self.setRight(self.right.rotateRight());
            }
            self = self.rotateLeft();
        }
        while (self.isLeftOriented()) {
            if (self.left.isRightOriented()) {
                self.setLeft(self.left.rotateLeft());
            }
            self = self.rotateRight();
        }
        return self;
    }
    // Traversal
    min() {
        if (this.left !== undefined) {
            return this.left.min(); // tail recursion
        }
        return this.value;
    }
    max() {
        if (this.right !== undefined) {
            return this.right.max(); // tail recursion
        }
        return this.value;
    }
    predecessor() {
        if (this.left !== undefined) {
            return this.left.max();
        }
        return undefined;
    }
    successor() {
        if (this.right !== undefined) {
            return this.right.min();
        }
        return undefined;
    }
    // Removal
    removeMin() {
        if (this.left !== undefined) {
            this.setLeft(this.left.removeMin());
            return this.balance();
        }
        return this.right;
    }
    removeMax() {
        if (this.right !== undefined) {
            this.setRight(this.right.removeMax());
            return this.balance();
        }
        return this.left;
    }
    /** @Override */
    removePredecessor() {
        if (this.left !== undefined) {
            this.setLeft(this.left.removeMax());
        }
        // delegate balance to parent
    }
    /** @Override */
    removeSuccessor() {
        if (this.right !== undefined) {
            this.setRight(this.right.removeMin());
        }
        // delegate balance to parent
    }
    // merge
    mergeLeft() {
        const curr = this.value;
        const prev = this.predecessor();
        if (prev !== undefined &&
            prev.compare(curr) === -5 /* PREPENDABLE */) {
            this.removePredecessor();
            this.value = prev.append(curr);
        }
    }
    mergeRight() {
        const curr = this.value;
        const succ = this.successor();
        if (succ !== undefined &&
            succ.compare(curr) === 5 /* APPENDABLE */) {
            this.removeSuccessor();
            this.value = curr.append(succ);
        }
    }
    // insertable
    insertableLeft(iRange) {
        if (this.left !== undefined) {
            return this.left.insertable(iRange);
        }
        return [iRange];
    }
    insertableRight(iRange) {
        if (this.right !== undefined) {
            return this.right.insertable(iRange);
        }
        return [iRange];
    }
    insertable(iRange) {
        const curr = this.value;
        switch (iRange.compare(curr)) {
            case -5 /* PREPENDABLE */:
            case -6 /* BEFORE */:
                return this.insertableLeft(iRange);
            case 5 /* APPENDABLE */:
            case 6 /* AFTER */:
                return this.insertableRight(iRange);
            case 3 /* INCLUDING_RIGHT */:
            case -4 /* OVERLAPPING_BEFORE */:
                return this.insertableLeft(iRange.prependable(curr));
            case 1 /* INCLUDING_LEFT */:
            case 4 /* OVERLAPPING_AFTER */:
                return this.insertableRight(iRange.appendable(curr));
            case 2 /* INCLUDING_MIDDLE */: {
                const lIns = this.insertableLeft(iRange.prependable(curr));
                const rIns = this.insertableRight(iRange.appendable(curr));
                return lIns.concat(rIns);
            }
            case -2 /* INCLUDED_MIDDLE_BY */:
            case -1 /* INCLUDED_RIGHT_BY */:
            case -3 /* INCLUDED_LEFT_BY */:
            case 0 /* EQUAL */:
                return [];
        }
    }
    // insert
    insertLeft(iRange) {
        let ins;
        if (this.left !== undefined) {
            ins = this.left.insert(iRange);
            this.setLeft(this.left.balance());
        }
        else {
            this.setLeft(ValuedNode.leaf(iRange));
            ins = [iRange];
        }
        // delegate balance to parent
        return ins;
    }
    insertRight(iRange) {
        let ins;
        if (this.right !== undefined) {
            ins = this.right.insert(iRange);
            this.setRight(this.right.balance());
        }
        else {
            // should not happen
            this.setRight(ValuedNode.leaf(iRange));
            ins = [iRange];
        }
        // delegate balance to parent
        return ins;
    }
    insert(iBlock) {
        const curr = this.value;
        switch (iBlock.compare(curr)) {
            case -6 /* BEFORE */:
                return this.insertLeft(iBlock);
            case 6 /* AFTER */:
                return this.insertRight(iBlock);
            case -5 /* PREPENDABLE */: {
                const lIns = this.insertLeft(iBlock);
                this.mergeLeft();
                return lIns;
            }
            case 5 /* APPENDABLE */: {
                const rIns = this.insertRight(iBlock);
                this.mergeRight();
                return rIns;
            }
            case -4 /* OVERLAPPING_BEFORE */:
            case 4 /* OVERLAPPING_AFTER */:
            case -2 /* INCLUDED_MIDDLE_BY */:
            case -1 /* INCLUDED_RIGHT_BY */:
            case -3 /* INCLUDED_LEFT_BY */:
            case 2 /* INCLUDING_MIDDLE */:
            case 3 /* INCLUDING_RIGHT */:
            case 1 /* INCLUDING_LEFT */:
            case 0 /* EQUAL */:
                throw new Error("inserted value intersects with one or more existing blocks");
        }
    }
}
//# sourceMappingURL=causal-context-node.js.map